@@ -13,7 +13,7 @@ import datetime
 import logging
 from dataclasses import dataclass, field
 
-from typing import Coroutine, Literal, NoReturn, overload
+from typing import Any, Callable, Coroutine, Literal, NoReturn, overload
 
 import sshkeyboard
 from pydantic import BaseModel, field_serializer
@@ -25,11 +25,19 @@ from rich.panel import Panel
 
 from lvmopstools.devices.specs import spectrograph_pressures, spectrograph_temperatures
 from lvmopstools.devices.thermistors import read_thermistors
+from lvmopstools.retrier import Retrier
+from sdsstools.utils import GatheringTaskGroup
 
 from lvmcryo.config import ValveConfig, get_internal_config
 from lvmcryo.handlers.thermistor import ThermistorMonitor
 from lvmcryo.handlers.valve import ValveHandler
-from lvmcryo.tools import TimerProgressBar, cancel_task, get_fake_logger, o2_alert
+from lvmcryo.tools import (
+    TimerProgressBar,
+    cancel_task,
+    get_fake_logger,
+    ln2_estops,
+    o2_alert,
+)
 
 
 def convert_datetime_to_iso_8601_with_z_suffix(dt: datetime.datetime) -> str:
@@ -181,6 +189,8 @@ class LN2Handler:
 
         """
 
+        retrier = Retrier(max_attempts=3, delay=1)
+
         # Check O2 alarms.
         if not self.alerts_route:
             self.log.warning("No alerts route provided. Not checking O2 alarms.")
@@ -199,14 +209,14 @@ class LN2Handler:
             self.log.info("Checking connection to NPS outlets ...")
             for valve in self.valve_handlers:
                 if not await self.valve_handlers[valve].check():
-                    raise RuntimeError(f"valve {valve} failed or did not reply.")
+                    raise RuntimeError(f"valve {valve!r} failed or did not reply.")
         except Exception as err:
             self.fail(f"Failed checking connection to NPS outlets: {err}")
 
         if max_temperature is not None:
             self.log.info("Checking LN2 temperatures ...")
             try:
-                spec_temperatures = await spectrograph_temperatures()
+                spec_temperatures = await retrier(spectrograph_temperatures)()
             except Exception as err:
                 self.fail(f"Failed reading spectrograph temperatures: {err}")
             else:
@@ -218,15 +228,15 @@ class LN2Handler:
 
                     if ln2_temp > max_temperature:
                         self.fail(
-                            f"LN2 temperature for camera {camera} is {ln2_temp:.1f} C "
-                            f"which is above the maximum allowed temperature "
-                            f"({max_temperature:.1f} C)."
+                            f"LN2 temperature for camera {camera!r} is "
+                            f"{ln2_temp:.1f} C which is above the maximum allowed "
+                            f"temperature ({max_temperature:.1f} C)."
                         )
 
         if max_pressure is not None:
             self.log.info("Checking pressures ...")
             try:
-                spec_pressures = await spectrograph_pressures()
+                spec_pressures = await retrier(spectrograph_pressures)()
             except Exception as err:
                 self.fail(f"Failed reading spectrograph pressures: {err}")
             else:
@@ -238,7 +248,7 @@ class LN2Handler:
 
                     if pressure > max_pressure:
                         self.fail(
-                            f"Pressure for camera {camera} is {pressure} Torr "
+                            f"Pressure for camera {camera!r} is {pressure} Torr "
                             f"which is above the maximum allowed pressure "
                             f"({max_pressure} Torr)."
                         )
@@ -255,19 +265,19 @@ class LN2Handler:
                 thermistor_info = self.valve_info[valve].thermistor
 
                 if thermistor_info is None:
-                    self.log.warning(f"Cannot check thermistor for {valve}.")
+                    self.log.warning(f"Cannot check thermistor for {valve!r}.")
                     continue
 
                 if thermistor_info.disabled:
-                    self.log.warning(f"Thermistor for {valve} is disabled.")
+                    self.log.warning(f"Thermistor for {valve!r} is disabled.")
                     continue
 
                 if thermistor_info.channel is None:
-                    self.fail(f"Thermistor channel for {valve} not defined.")
+                    self.fail(f"Thermistor channel for {valve!r} not defined.")
 
                 thermistor_value = thermistors[thermistor_info.channel]
                 if thermistor_value is True:
-                    self.fail(f"Thermistor for valve {valve} is active.")
+                    self.fail(f"Thermistor for valve {valve!r} is active.")
 
         self.log.info("All pre-fill checks passed.")
 
@@ -280,6 +290,7 @@ class LN2Handler:
         min_purge_time: float | None = None,
         max_purge_time: float | None = None,
         prompt: bool | None = None,
+        preopen_cb: Callable[[], Coroutine | Any] | None = None,
     ):
         """Purges the system.
 
@@ -300,6 +311,8 @@ class LN2Handler:
         prompt
             Whether to show a prompt to stop or cancel the purge. If ``None``,
             determined from the instance ``interactive`` attribute.
+        preopen_cb
+            A callback to run before opening the purge valve.
 
         """
 
@@ -321,6 +334,12 @@ class LN2Handler:
             self._kb_monitor(action="purge")
 
         try:
+            if preopen_cb:
+                if asyncio.iscoroutinefunction(preopen_cb):
+                    await preopen_cb()
+                else:
+                    preopen_cb()
+
             await valve_handler.start_fill(
                 min_open_time=min_purge_time or 0.0,
                 max_open_time=max_purge_time,
@@ -347,9 +366,11 @@ class LN2Handler:
         self,
         cameras: list[str] | None = None,
         use_thermistors: bool = True,
+        require_all_thermistors: bool = False,
         min_fill_time: float | None = None,
         max_fill_time: float | None = None,
         prompt: bool | None = None,
+        preopen_cb: Callable[[], Coroutine | Any] | None = None,
     ):
         """Fills the selected cameras.
 
@@ -360,6 +381,9 @@ class LN2Handler:
             used to instantiate the `.LN2Handler` instance.
         use_thermistors
             Whether to use the thermistors to close the valves.
+        require_all_thermistors
+            Whether to require all thermistors to be active before closing the
+            valves. If `False`, valves as closed as their thermistors become active.
         min_fill_time
             The minimum time to keep the fill valves open. Only relevant if
             using the thermistors.
@@ -371,6 +395,8 @@ class LN2Handler:
         prompt
             Whether to show a prompt to stop or cancel the fill. If ``None``,
             determined from the instance ``interactive`` attribute.
+        preopen_cb
+            A callback to run before opening the fill valves.
 
         """
 
@@ -391,6 +417,7 @@ class LN2Handler:
                     min_open_time=min_fill_time or 0.0,
                     max_open_time=max_fill_time,
                     use_thermistor=use_thermistors,
+                    close_on_active=not require_all_thermistors,
                 )
             )
 
@@ -407,6 +434,12 @@ class LN2Handler:
             self._kb_monitor(action="fill")
 
         try:
+            if preopen_cb:
+                if asyncio.iscoroutinefunction(preopen_cb):
+                    await preopen_cb()
+                else:
+                    preopen_cb()
+
             await asyncio.gather(*fill_tasks)
 
             if not self.aborted:
@@ -417,6 +450,14 @@ class LN2Handler:
             raise
 
         finally:
+            if use_thermistors and require_all_thermistors:
+                # If we are waiting for all thermistors to be active,
+                # we need to close all valves now.
+                self.log.info("Closing all valves.")
+                async with GatheringTaskGroup() as group:
+                    for valve_handler in self.valve_handlers.values():
+                        group.create_task(valve_handler._set_state(False))
+
             self.event_times.fill_complete = get_now()
             if prompt:
                 sshkeyboard.stop_listening()
@@ -486,7 +527,7 @@ class LN2Handler:
 
             elif key == "enter":
                 self.log.warning("Finishing purge/fill.")
-                await self.close_valves(only_active=True)
+                await self.stop(only_active=True)
 
         self.console.print(
             Panel(
@@ -535,6 +576,21 @@ class LN2Handler:
                 self.log.warning(f"Error reading alerts: {ee}")
                 n_failed += 1
 
+            try:
+                ln2_estops_active = await ln2_estops()
+                if ln2_estops_active:
+                    await self.abort(
+                        error="LN2 estops detected: aborting.",
+                        close_valves=False,  # The NPS will be unavailable
+                        raise_error=False,
+                    )
+            except Exception:
+                # Log the error but do not increase n_failed. The e-stops are a
+                # low-level safety feature. We only monitor them here to cleanly
+                # abort the fill if they are pressed, but worst case the code
+                # will fail but that won't have an impact on the system.
+                self.log.warning("Error reading LN2 estops.")
+
             if n_failed >= 10:
                 await self.abort(
                     error="Too many errors reading alerts. Aborting.",
@@ -544,7 +600,7 @@ class LN2Handler:
 
             await asyncio.sleep(3)
 
-    async def close_valves(self, only_active: bool = True):
+    async def stop(self, close_valves: bool = True, only_active: bool = True):
         """Cancels ongoing fills and closes the valves.
 
         If ``only_active=True`` only active valves will be closed. Otherwise
@@ -556,9 +612,14 @@ class LN2Handler:
 
         for valve_handler in self.valve_handlers.values():
             if valve_handler.active or not only_active:
-                tasks.append(valve_handler.finish())
+                tasks.append(valve_handler.finish(close_valve=close_valves))
 
-        await asyncio.gather(*tasks)
+        # Try to close as many valves as possible but then raise the error if
+        # any failed.
+        results = await asyncio.gather(*tasks, return_exceptions=True)
+        for result in results:
+            if isinstance(result, Exception):
+                raise result
 
     async def clear(self):
         """Cleanly finishes tasks and other clean-up tasks."""
@@ -627,7 +688,7 @@ class LN2Handler:
         self.aborted = True
         self.event_times.abort_time = get_now()
 
-        if close_valves:
-            await self.close_valves(only_active=False)
+        # For each valve, close it (optionally) and finish the monitoring.
+        await self.stop(close_valves=close_valves, only_active=False)
 
         self.fail(error=error or "Aborted.", raise_error=raise_error)
